use anyhow::{Context, Result, bail};
use serde_json::Value;
use std::collections::{HashMap, HashSet};
use std::fs;
use tracing::{info, instrument};

pub mod cli;
pub mod downloader;

use cli::Args;

/// The main application logic.
#[instrument(skip_all, fields(
    input_config = %args.input_config.display(),
    rules_dir = %args.rules_dir.display(),
    output_config = %args.output_config.display()
))]
pub async fn run(args: Args) -> Result<()> {
    info!("Starting ruleset synchronization");

    // 1. Read the input config file into a generic, order-preserving Value.
    // We only use this to discover the URLs, not to modify.
    let config_content = fs::read_to_string(&args.input_config).with_context(|| {
        format!(
            "Failed to load input config from {}",
            args.input_config.display()
        )
    })?;
    let config_value: Value = serde_json::from_str(&config_content)
        .context("Failed to parse JSON. Please check the config file structure.")?;

    // 2. Build the list of URLs to download, the list of expected files,
    // AND a map of (original_url -> new_url) for text replacement later.
    let mut urls_to_download = Vec::new();
    let mut expected_files = HashSet::new();
    let mut url_replacement_map = HashMap::new();

    if let Some(rule_sets) = config_value
        .pointer("/route/rule_set")
        .and_then(|v| v.as_array())
    {
        let clean_domain = args.domain.trim_matches('/');
        let clean_rule_path = format!("/{}/", args.rule_path.trim_matches('/'));

        for rule in rule_sets {
            if let Some(url) = rule.get("url").and_then(|u| u.as_str()) {
                urls_to_download.push(url.to_string());
                if let Ok(filename) = downloader::url_to_filename(url) {
                    expected_files.insert(args.rules_dir.join(&filename));
                    // Create the new URL and store the mapping.
                    let new_url =
                        format!("https://{}{}{}", clean_domain, clean_rule_path, filename);
                    url_replacement_map.insert(url.to_string(), new_url);
                }
            }
        }
    }

    // 3. Ensure the rules directory exists and clean up stale files.
    fs::create_dir_all(&args.rules_dir).with_context(|| {
        format!(
            "Failed to create rules directory '{}'",
            args.rules_dir.display()
        )
    })?;
    downloader::cleanup_stale_files(&args.rules_dir, &expected_files)?;

    // 4. Proceed to download new/updated files.
    if urls_to_download.is_empty() {
        info!("No rule sets with URLs found in input config.");
    } else {
        info!(
            count = urls_to_download.len(),
            "Found rule sets to download/update."
        );
        let report =
            downloader::download_all_rules(&urls_to_download, &args.rules_dir, args.concurrency)
                .await?;
        info!(
            successful = report.successful,
            failed = report.failed,
            "Download process finished."
        );
        if report.failed > 0 {
            bail!(
                "Aborting due to {} download failures. Output config was NOT generated.",
                report.failed
            );
        }
    }

    // 5. Generate the new config file using simple text replacement.
    info!("Generating output configuration file...");

    // We use the original `config_content` string read at the start.
    let mut output_content = config_content;

    // Perform a direct search-and-replace for each URL.
    for (original_url, new_url) in &url_replacement_map {
        output_content = output_content.replace(original_url, new_url);
    }

    // Write the modified string content to the output file.
    fs::write(&args.output_config, output_content).with_context(|| {
        format!(
            "Failed to write output config to {}",
            args.output_config.display()
        )
    })?;
    info!(
        path = %args.output_config.display(),
        "Output configuration file saved successfully."
    );

    info!("Synchronization process completed successfully.");
    Ok(())
}
